# This code is provided without warranty expressed nor implied.

import random

# The set of RPN functions
# the stack is always reversed order
def rpn_negate(stack):
    stack.append(-stack.pop())


def rpn_xtoy(stack):
    y = stack.pop()
    x = stack.pop()
    stack.append(x ** y)


def rpn_shift_right(stack):
    i = round(stack.pop(), 0)
    x = round(stack.pop(), 0)
    for idx in range(int(i)):
        x *= 2
    stack.append(x)


def rpn_shift_left(stack):
    i = round(stack.pop(), 0)
    x = round(stack.pop(), 0)
    for idx in range(int(i)):
        x /= 2
    stack.append(x)


def rpn_multiply(stack):
    stack.append(stack.pop() * stack.pop())


def rpn_divide(stack):
    y = stack.pop()
    x = stack.pop()
    stack.append(y / x)


def rpn_modulo(stack):
    y = stack.pop()
    x = stack.pop()
    stack.append(y % x)


def rpn_add(stack):
    stack.append(stack.pop() + stack.pop())


def rpn_subtract(stack):
    y = stack.pop()
    x = stack.pop()
    stack.append(y - x)


def rpn_abs(stack):
    stack.append(abs(stack.pop()))


def rpn_round(stack):
    r = stack.pop()
    x = stack.pop()
    stack.append(round(x, int(round(r, 0))))


def rpn_random(stack):
    stack.append(random.random())


# This is a set of operators, the number of arguments and the function
# The arguments are used for the optimizer to work.  It must match the actual
# destacking in the function
op = {'(': {'precedence': 99, 'arguments': 0, 'function': None},
      '!': {'precedence': 70, 'arguments': 1, 'function': rpn_negate},
      '^': {'precedence': 60, 'arguments': 2, 'function': rpn_xtoy},
      '<<': {'precedence': 60, 'arguments': 2, 'function': rpn_shift_right},
      '>>': {'precedence': 60, 'arguments': 2, 'function': rpn_shift_left},
      '*': {'precedence': 50, 'arguments': 2, 'function': rpn_multiply},
      '/': {'precedence': 50, 'arguments': 2, 'function': rpn_divide},
      '%': {'precedence': 50, 'arguments': 2, 'function': rpn_modulo},
      '+': {'precedence': 40, 'arguments': 2, 'function': rpn_add},
      '-': {'precedence': 40, 'arguments': 2, 'function': rpn_subtract},
      ',': {'precedence': 0, 'arguments': 1, 'function': None}
      }

# These are the functions
# The arguments are also there for the optimizer but also for syntax chacking with the correct
# parenthesis closures and commas.  The optimize member is there so that certain functions do not get
# optimized out such as the random function
function_priority = 90
func = {'ABS': {'precedence': function_priority, 'arguments': 1, 'function': rpn_abs, 'optimize': True},
        'ROUND': {'precedence': function_priority, 'arguments': 2, 'function': rpn_round, 'optimize': True},
        'RND': {'precedence': function_priority, 'arguments': 0, 'function': rpn_random, 'optimize': False}
        }

# These are variables that are inputs to the equation.
var = ['A', 'B', 'C', 'D']

# The error codes
no_error = 0
invalid_operand = 1
invalid_operator = 2
incorrect_parenthesis = 3
mismatch_parenthesis = 4
too_few_arguments = 5
too_many_arguments = 6

# Returns the string for a specific error
def get_error_string(error):
    if error == no_error:
        return 'No Error'
    elif error == invalid_operand:
        return "Invalid Operand"
    elif error == invalid_operator:
        return "Invalid Operand"
    elif error == incorrect_parenthesis:
        return "Incorrect Parenthesis"
    elif error == mismatch_parenthesis:
        return "Mismatch Parenthesis"
    elif error == too_few_arguments:
        return "Too Few Arguments"
    elif error == too_many_arguments:
        return "Too Many Arguments"
    else:
        return "Unknown: " + str(error)


# Returns an operator for the string c including multi character operators
def get_operator(c):
    if c[0] == ')':
        return ')'

    for operator in op:
        c2 = c[:len(operator)]
        if c2 == operator:
            return operator
    return None

# This class represents a function, operator, constant or variable
# It also represents exactly one stack entity
class StackItem:
    def __init__(self, value, offset, is_operator=False, is_close_paren=False):
        self.value = value                          # the string value
        self.offset = offset                        # the offset into the original infix string for error reporting
        self.is_operator = is_operator              # true if an operator
        self.is_close_paren = is_close_paren        # true if a close parenthesis
        self.precedence = 0                         # the precedence
        self.is_function = False                    # true if a function
        self.is_constant = False                    # true if a constant
        self.is_variable = False                    # true if a variable
        self.function = None                        # the function
        self.arguments = 0                          # the number of arguments for the function
        self.optimize = True                        # can this function be optimized
        self.error = no_error                       # the error

    # this method is useful when a subtraction sign is determined to be a negation
    def negate(self):
        if self.value == '-':
            self.value = '!'
            item = op[self.value]
            self.precedence = item['precedence']
            self.function = item['function']
            self.arguments = item['arguments']

    # used to determine if this stack item is a constant
    def evaluate_constant(self):
        for idx in range(len(self.value)):
            c = self.value[idx]
            if c not in '0123456789.':
                return False
        return True

    # evaluates the type of stack item
    # if this isn't a close parenthesis, then determine its type
    def evaluate(self, op, func, var):
        if self.is_close_paren:
            return True

        if self.is_operator:
            item = op[self.value]
            self.precedence = item['precedence']
            self.function = item['function']
            self.arguments = item['arguments']
            return True

        if self.value in func:
            self.is_function = True
            item = func[self.value]
            self.precedence = item['precedence']
            self.function = item['function']
            self.arguments = item['arguments']
            self.optimize = item['optimize']
            return True

        if self.value in var:
            self.is_variable = True
            return True

        if self.evaluate_constant():
            self.is_constant = True
            return True

        self.error = invalid_operand
        return False

    # returns the type of the stack item
    def get_type_string(self):
        if self.is_operator:
            return 'operator'

        if self.is_close_paren:
            return 'close'

        if self.is_function:
            return 'function'

        if self.is_variable:
            return 'variable'

        if self.is_constant:
            return 'constant'

        if self.error:
            return 'error:' + str(self.offset)

        return 'unknown'

    # string representation
    def __repr__(self):
        return '<' + self.value + ', ' + self.get_type_string() + '>'


# Step 1 - convert the infix string to stack items
def infix_to_stack(infix):
    # uppercase and remove space
    infix = infix.upper()

    stack = []
    last_idx = 0
    end_of_infix = len(infix)
    idx = 0
    # iterate through each character of the infix string and determine which are operators
    while idx < end_of_infix:
        c = infix[idx:]
        operator = get_operator(c)
        if operator is not None:
            # we found an operator so stack the info since the last operator
            text = infix[last_idx:idx].rstrip()
            orig_len = len(text)
            text = text.lstrip()
            last_idx += orig_len - len(text)
            if len(text) > 0:
                stack.append(StackItem(text, last_idx))

            # Now stack the operator or the close parenthesis
            if operator == ')':
                stack.append(StackItem(operator, idx, False, True))
            else:
                stack.append(StackItem(operator, idx, True))
            idx += len(operator)
            last_idx = idx
        else:
            idx += 1

    # if there are any remaining information, then stack it as text
    if last_idx <= idx:
        text = infix[last_idx:].rstrip()
        orig_len = len(text)
        text = text.lstrip()
        last_idx += orig_len - len(text)
        if len(text) > 0:
            stack.append(StackItem(text, last_idx))

    # Next evalute every stack item to determine the type
    for si in stack:
        si.evaluate(op, func, var)

    # return an stacked infix representation
    return stack


# Step 2 - determine if the '-' is a subtraction or a negation
# the item before must be an operator.  The item after must be a constant or variable.
# if so, it is a negation and not a subtraction so change the stack item's type
def alter_negation(stack):
    for idx in range(1, len(stack) - 1):
        if stack[idx].value == '-' and stack[idx - 1].value in op and \
                (stack[idx + 1].is_constant or stack[idx + 1].is_variable or stack[idx + 1].is_function):
            stack[idx].negate()
    return stack


# Step 3 - make sure no operators are back to back except for negation, parenthesis, and comma
def check_operators(stack):
    stack_wo_negates = [si for si in stack if si.value not in '!(),']
    for si in stack_wo_negates:
        if si.is_operator:
            op_cnt += 1
        else:
            op_cnt = 0
        if op_cnt > 1:
            si.error = invalid_operator
    return stack

# Step 4 - check for mismatch parenthesis
def check_mismatch_paren(stack):
    paren_cnt = 0;
    for idx in range(len(stack)):
        si = stack[idx]
        if si.value == '(':
            paren_cnt += 1
        if si.value == ')':
            paren_cnt -= 1
            if paren_cnt < 0:
                si.error = mismatch_parenthesis
            elif (idx > 0) and stack[idx - 1].is_operator:
                si.error = incorrect_parenthesis

    if paren_cnt != 0:
        stack[-1].error = mismatch_parenthesis;
    return stack

# Step 5 - check to make sure the functions have the correct number of arguments
def check_functions(stack):
    func_list = []
    end_of_stack = len(stack)
    for idx in range(end_of_stack):
        si = stack[idx]
        # if we are evaluating a function
        if si.is_function:
            arguments = func[si.value]['arguments']

            # if no arguments, then no parenthesis
            if arguments == 0:
                if idx < (end_of_stack - 1):
                    if stack[idx + 1].value == '(':
                        stack[idx + 1].error = incorrect_parenthesis
                        return stack
            # other wise stack of the function work through the arguments and commas
            else:
                func_list.append({'comma_cnt': func[si.value]['arguments'] - 1, 'paren_cnt': 0})
                if idx < (end_of_stack - 1):
                    if stack[idx + 1].value != '(':
                        stack[idx + 1].error = invalid_operator
                        return stack

        # if we are not in a function, then there should be no commas
        elif len(func_list) == 0:
            if si.value == ',':
                si.error = invalid_operator
                return stack

        # if we are evaluating a function, count the parenthsis and the commas
        else:
            last_func = func_list[-1]
            if si.value == '(':
                last_func['paren_cnt'] += 1
            if si.value == ')':
                if last_func['comma_cnt'] == 0:
                    func_list.pop()
                else:
                    si.error = too_few_arguments
                    return stack
            if si.value == ',':
                last_func['comma_cnt'] -= 1
                if last_func['comma_cnt'] < 0:
                    si.error = too_many_arguments
                    return stack
    return stack


# Convert the infix to postfix
def shunt(stack):
    postfix = []
    op_stack = []

    for si in stack:
        if si.is_variable or si.is_constant:
            postfix.append(si)
        elif si.value == '(':
            op_stack.append(si)
        elif si.is_close_paren:
            si_op = op_stack.pop()
            while si_op.value != '(':
                postfix.append(si_op)
                si_op = op_stack.pop()
        else:  # operators and functions
            while len(op_stack) > 0:
                last_op = op_stack[-1]
                if (last_op.value != '(') and (last_op.precedence >= si.precedence):
                    postfix.append(op_stack.pop())
                else:
                    break
            op_stack.append(si)

    while len(op_stack) > 0:
        postfix.append(op_stack.pop())

    # remove commas now because they have no need in postfix
    postfix = [si for si in postfix if si.value != ',']

    return postfix, stack

# this runs through the stack create a new stack where functions and arguments of only constants get reduced
def optimize(postfix):
    if len(postfix) == 0:
        return postfix

    opt_postfix = []

    for si in postfix:
        if (si.is_operator or si.is_function) and si.optimize:
            rpn_stack = []
            all_constants = True

            for idx in range(si.arguments):
                if not opt_postfix[-(idx + 1)].is_constant:
                    all_constants = False
                    break

            if all_constants:
                for i in range(si.arguments):
                    si_op = opt_postfix.pop()
                    rpn_stack.insert(0, float(si_op.value))
                si.function(rpn_stack)
                new_si = StackItem(str(rpn_stack.pop()), si.offset)
                new_si.is_constant = True
                opt_postfix.append(new_si)
                continue

        opt_postfix.append(si)

    return opt_postfix


# execute the RPN stack with the variables as inputs
def execute(variables, postfix):
    if len(postfix) == 0:
        return 0.0

    rpn_stack = []

    for si in postfix:
        if si.is_constant:
            rpn_stack.append(float(si.value))

        elif si.is_variable:
            rpn_stack.append(float(variables[si.value]))

        elif si.is_operator or si.is_function:
            si.function(rpn_stack)

    return rpn_stack.pop()


def doit(infix):
    print('------------------------------------------------')
    print(infix)

    # perform the steps
    stack = infix_to_stack(infix)

    print(stack)

    stack = alter_negation(stack)

    print(stack)

    stack = check_operators(stack)
    stack = check_mismatch_paren(stack)
    stack = check_functions(stack)

    # print out any errors
    error_list = [si for si in stack if si.error != no_error]
    if len(error_list) > 0:
        for si in error_list:
            if si.error != no_error:
                print(infix)
                print('^'.rjust(si.offset + 1))
                print('Error: ', get_error_string(si.error))
    else:
        variables = {'A': 3.42, 'B': 18, 'C': 42}

        # shut int
        postfix, stack = shunt(stack)

        # execute for fun
        print(postfix)
        result = execute(variables, postfix)
        print(result)

        # optimize
        postfix = optimize(postfix)

        # execute for fun again
        print(postfix)
        result = execute(variables, postfix)
        print(result)

        # for byte ish verions, compile to byte code and constants list


# doit('A * -3A2 + ABS(C) * D')
# doit('AB * -3A2 + ABS(C) * D')
# doit('AB * -3A2 + FRED(C) * D')
# doit('A * 32 + C * D')
# doit('A * / -32.834 + ABS(-C) * D')
# doit('A *  / - 32.834 + ABS(-C) * D')
# doit('A*32+C*D')
# doit('A*(32+C)*D')
# doit('A * -32.834 + ABS(-B) * C')
# doit('A * (-32.834 +  ABS(-B)) * C')
# doit('A * -32.834')
# doit('1 << 3')
# doit('8 >> 3')
doit('2 * (1 + -3) + abs(-round(rnd, 2))')
